@@ -1,5 +1,5 @@
 /* Reading NeXTstep/GNUstep .strings files.
-   Copyright (C) 2003, 2005-2007, 2009, 2019 Free Software Foundation, Inc.
+   Copyright (C) 2003-2024 Free Software Foundation, Inc.
    Written by Bruno Haible <bruno@clisp.org>, 2003.
 
    This program is free software: you can redistribute it and/or modify
@@ -29,12 +29,13 @@
 #include <stdlib.h>
 #include <string.h>
 
-#include "error.h"
-#include "error-progname.h"
+#include <error.h>
+#include "attribute.h"
 #include "read-catalog-abstract.h"
 #include "xalloc.h"
 #include "xvasprintf.h"
-#include "po-xerror.h"
+#include "xstrerror.h"
+#include "xerror-handler.h"
 #include "unistr.h"
 #include "gettext.h"
 
@@ -66,7 +67,7 @@
 static const char *real_file_name;
 
 /* File name and line number.  */
-extern lex_pos_ty gram_pos;
+static lex_pos_ty pos;
 
 /* The input file stream.  */
 static FILE *fp;
@@ -79,7 +80,7 @@ static unsigned char phase1_pushback[4];
 static int phase1_pushback_length;
 
 static int
-phase1_getc ()
+phase1_getc (abstract_catalog_reader_ty *catr)
 {
   int c;
 
@@ -92,12 +93,11 @@ phase1_getc ()
     {
       if (ferror (fp))
         {
-          const char *errno_description = strerror (errno);
-          po_xerror (PO_SEVERITY_FATAL_ERROR, NULL, NULL, 0, 0, false,
-                     xasprintf ("%s: %s",
-                                xasprintf (_("error while reading \"%s\""),
-                                           real_file_name),
-                                errno_description));
+          int err = errno;
+          catr->xeh->xerror (CAT_SEVERITY_FATAL_ERROR, NULL, NULL, 0, 0, false,
+                             xstrerror (xasprintf (_("error while reading \"%s\""),
+                                                   real_file_name),
+                                        err));
         }
       return EOF;
     }
@@ -137,7 +137,7 @@ enum enc
 static enum enc encoding;
 
 static int
-phase2_getc ()
+phase2_getc (abstract_catalog_reader_ty *catr)
 {
   if (phase2_pushback_length)
     return phase2_pushback[--phase2_pushback_length];
@@ -147,10 +147,10 @@ phase2_getc ()
       /* Determine the input file's encoding.  */
       int c0, c1;
 
-      c0 = phase1_getc ();
+      c0 = phase1_getc (catr);
       if (c0 == EOF)
         return UEOF;
-      c1 = phase1_getc ();
+      c1 = phase1_getc (catr);
       if (c1 == EOF)
         {
           phase1_ungetc (c0);
@@ -164,7 +164,7 @@ phase2_getc ()
         {
           int c2;
 
-          c2 = phase1_getc ();
+          c2 = phase1_getc (catr);
           if (c2 == EOF)
             {
               phase1_ungetc (c1);
@@ -190,10 +190,10 @@ phase2_getc ()
       {
         int c0, c1;
 
-        c0 = phase1_getc ();
+        c0 = phase1_getc (catr);
         if (c0 == EOF)
           return UEOF;
-        c1 = phase1_getc ();
+        c1 = phase1_getc (catr);
         if (c1 == EOF)
           return UEOF;
         return (c0 << 8) + c1;
@@ -204,10 +204,10 @@ phase2_getc ()
       {
         int c0, c1;
 
-        c0 = phase1_getc ();
+        c0 = phase1_getc (catr);
         if (c0 == EOF)
           return UEOF;
-        c1 = phase1_getc ();
+        c1 = phase1_getc (catr);
         if (c1 == EOF)
           return UEOF;
         return c0 + (c1 << 8);
@@ -221,7 +221,7 @@ phase2_getc ()
         int c;
         ucs4_t uc;
 
-        c = phase1_getc ();
+        c = phase1_getc (catr);
         if (c == EOF)
           return UEOF;
         buf[0] = c;
@@ -229,7 +229,7 @@ phase2_getc ()
 
         if (buf[0] >= 0xc0)
           {
-            c = phase1_getc ();
+            c = phase1_getc (catr);
             if (c == EOF)
               return UEOF;
             buf[1] = c;
@@ -238,7 +238,7 @@ phase2_getc ()
             if (buf[0] >= 0xe0
                 && ((buf[1] ^ 0x80) < 0x40))
               {
-                c = phase1_getc ();
+                c = phase1_getc (catr);
                 if (c == EOF)
                   return UEOF;
                 buf[2] = c;
@@ -247,7 +247,7 @@ phase2_getc ()
                 if (buf[0] >= 0xf0
                     && ((buf[2] ^ 0x80) < 0x40))
                   {
-                    c = phase1_getc ();
+                    c = phase1_getc (catr);
                     if (c == EOF)
                       return UEOF;
                     buf[3] = c;
@@ -256,7 +256,7 @@ phase2_getc ()
                     if (buf[0] >= 0xf8
                         && ((buf[3] ^ 0x80) < 0x40))
                       {
-                        c = phase1_getc ();
+                        c = phase1_getc (catr);
                         if (c == EOF)
                           return UEOF;
                         buf[4] = c;
@@ -265,7 +265,7 @@ phase2_getc ()
                         if (buf[0] >= 0xfc
                             && ((buf[4] ^ 0x80) < 0x40))
                           {
-                            c = phase1_getc ();
+                            c = phase1_getc (catr);
                             if (c == EOF)
                               return UEOF;
                             buf[5] = c;
@@ -283,7 +283,7 @@ phase2_getc ()
     case enc_iso8859_1:
       /* Read an ISO-8859-1 encoded character.  */
       {
-        int c = phase1_getc ();
+        int c = phase1_getc (catr);
 
         if (c == EOF)
           return UEOF;
@@ -306,12 +306,12 @@ phase2_ungetc (int c)
 /* Phase 3: Read an UCS-4 character, with line number handling.  */
 
 static int
-phase3_getc ()
+phase3_getc (abstract_catalog_reader_ty *catr)
 {
-  int c = phase2_getc ();
+  int c = phase2_getc (catr);
 
   if (c == '\n')
-    gram_pos.line_number++;
+    pos.line_number++;
 
   return c;
 }
@@ -320,7 +320,7 @@ static void
 phase3_ungetc (int c)
 {
   if (c == '\n')
-    --gram_pos.line_number;
+    --pos.line_number;
   phase2_ungetc (c);
 }
 
@@ -330,18 +330,18 @@ static char *
 conv_from_ucs4 (const int *buffer, size_t buflen)
 {
   unsigned char *utf8_string;
-  size_t pos;
+  size_t i;
   unsigned char *q;
 
   /* Each UCS-4 word needs 6 bytes at worst.  */
   utf8_string = XNMALLOC (6 * buflen + 1, unsigned char);
 
-  for (pos = 0, q = utf8_string; pos < buflen; )
+  for (i = 0, q = utf8_string; i < buflen; )
     {
       unsigned int uc;
       int n;
 
-      uc = buffer[pos++];
+      uc = buffer[i++];
       n = u8_uctomb (q, uc, 6);
       assert (n > 0);
       q += n;
@@ -473,11 +473,11 @@ special_comment_add (const char *flag)
 }
 
 static inline void
-special_comment_finish ()
+special_comment_finish (abstract_catalog_reader_ty *catr)
 {
   if (special_comment != NULL)
     {
-      po_callback_comment_special (special_comment);
+      catalog_reader_seen_comment_special (catr, special_comment);
       free (special_comment);
       special_comment = NULL;
     }
@@ -512,8 +512,9 @@ comment_add (int c)
   buffer[buflen++] = c;
 }
 
-static inline void
-comment_line_end (size_t chars_to_remove, bool test_for_fuzzy_msgstr)
+static void
+comment_line_end (abstract_catalog_reader_ty *catr,
+                  size_t chars_to_remove, bool test_for_fuzzy_msgstr)
 {
   char *line;
 
@@ -547,7 +548,7 @@ comment_line_end (size_t chars_to_remove, bool test_for_fuzzy_msgstr)
     special_comment_add (line + 6);
   else if (strlen (line) >= 9 && memcmp (line, "Comment: ", 9) == 0)
     /* A comment extracted from the source.  */
-    po_callback_comment_dot (line + 9);
+    catalog_reader_seen_comment_dot (catr, line + 9);
   else
     {
       char *last_colon;
@@ -561,10 +562,10 @@ comment_line_end (size_t chars_to_remove, bool test_for_fuzzy_msgstr)
         {
           /* A "File: <filename>:<number>" type comment.  */
           *last_colon = '\0';
-          po_callback_comment_filepos (line + 6, number);
+          catalog_reader_seen_comment_filepos (catr, line + 6, number);
         }
       else
-        po_callback_comment (line);
+        catalog_reader_seen_comment (catr, line);
     }
 }
 
@@ -573,14 +574,14 @@ comment_line_end (size_t chars_to_remove, bool test_for_fuzzy_msgstr)
    character.  */
 
 static int
-phase4_getc ()
+phase4_getc (abstract_catalog_reader_ty *catr)
 {
   int c;
 
-  c = phase3_getc ();
+  c = phase3_getc (catr);
   if (c != '/')
     return c;
-  c = phase3_getc ();
+  c = phase3_getc (catr);
   switch (c)
     {
     default:
@@ -601,7 +602,7 @@ phase4_getc ()
         /* Drop additional stars at the beginning of the comment.  */
         for (;;)
           {
-            c = phase3_getc ();
+            c = phase3_getc (catr);
             if (c != '*')
               break;
             last_was_star = true;
@@ -609,7 +610,7 @@ phase4_getc ()
         phase3_ungetc (c);
         for (;;)
           {
-            c = phase3_getc ();
+            c = phase3_getc (catr);
             if (c == UEOF)
               break;
             /* We skip all leading white space, but not EOLs.  */
@@ -619,7 +620,7 @@ phase4_getc ()
               {
               case '\n':
                 seen_newline = true;
-                comment_line_end (1, false);
+                comment_line_end (catr, 1, false);
                 comment_start ();
                 last_was_star = false;
                 trailing_stars = 0;
@@ -634,12 +635,12 @@ phase4_getc ()
                 if (last_was_star)
                   {
                     /* Drop additional stars at the end of the comment.  */
-                    comment_line_end (trailing_stars + 1,
+                    comment_line_end (catr, trailing_stars + 1,
                                       expect_fuzzy_msgstr_as_c_comment
                                       && !seen_newline);
                     break;
                   }
-                /* FALLTHROUGH */
+                FALLTHROUGH;
 
               default:
                 last_was_star = false;
@@ -656,14 +657,14 @@ phase4_getc ()
       comment_start ();
       for (;;)
         {
-          c = phase3_getc ();
+          c = phase3_getc (catr);
           if (c == '\n' || c == UEOF)
             break;
           /* We skip all leading white space, but not EOLs.  */
           if (!(buflen == 0 && (c == ' ' || c == '\t')))
             comment_add (c);
         }
-      comment_line_end (0, expect_fuzzy_msgstr_as_cxx_comment);
+      comment_line_end (catr, 0, expect_fuzzy_msgstr_as_cxx_comment);
       return '\n';
     }
 }
@@ -705,9 +706,9 @@ is_quotable (int c)
 
 /* Read a key or value string.
    Return the string in UTF-8 encoding, or NULL if no string is seen.
-   Return the start position of the string in *pos.  */
+   Return the start position of the string in *start_pos.  */
 static char *
-read_string (lex_pos_ty *pos)
+read_string (abstract_catalog_reader_ty *catr, lex_pos_ty *start_pos)
 {
   static int *buffer;
   static size_t bufmax;
@@ -716,26 +717,26 @@ read_string (lex_pos_ty *pos)
 
   /* Skip whitespace before the string.  */
   do
-    c = phase4_getc ();
+    c = phase4_getc (catr);
   while (is_whitespace (c));
 
   if (c == UEOF)
     /* No more string.  */
     return NULL;
 
-  *pos = gram_pos;
+  *start_pos = pos;
   buflen = 0;
   if (c == '"')
     {
       /* Read a string enclosed in double-quotes.  */
       for (;;)
         {
-          c = phase3_getc ();
+          c = phase3_getc (catr);
           if (c == UEOF || c == '"')
             break;
           if (c == '\\')
             {
-              c = phase3_getc ();
+              c = phase3_getc (catr);
               if (c == UEOF)
                 break;
               if (c >= '0' && c <= '7')
@@ -747,7 +748,7 @@ read_string (lex_pos_ty *pos)
                       n = n * 8 + (c - '0');
                       if (++j == 3)
                         break;
-                      c = phase3_getc ();
+                      c = phase3_getc (catr);
                       if (!(c >= '0' && c <= '7'))
                         {
                           phase3_ungetc (c);
@@ -762,7 +763,7 @@ read_string (lex_pos_ty *pos)
                   int j;
                   for (j = 0; j < 4; j++)
                     {
-                      c = phase3_getc ();
+                      c = phase3_getc (catr);
                       if (c >= '0' && c <= '9')
                         n = n * 16 + (c - '0');
                       else if (c >= 'A' && c <= 'F')
@@ -797,18 +798,18 @@ read_string (lex_pos_ty *pos)
           buffer[buflen++] = c;
         }
       if (c == UEOF)
-        po_xerror (PO_SEVERITY_ERROR, NULL,
-                   real_file_name, gram_pos.line_number, (size_t)(-1), false,
-                   _("warning: unterminated string"));
+        catr->xeh->xerror (CAT_SEVERITY_ERROR, NULL,
+                           real_file_name, pos.line_number, (size_t)(-1), false,
+                           _("warning: unterminated string"));
     }
   else
     {
       /* Read a token outside quotes.  */
       if (is_quotable (c))
-        po_xerror (PO_SEVERITY_ERROR, NULL,
-                   real_file_name, gram_pos.line_number, (size_t)(-1), false,
-                   _("warning: syntax error"));
-      for (; c != UEOF && !is_quotable (c); c = phase4_getc ())
+        catr->xeh->xerror (CAT_SEVERITY_ERROR, NULL,
+                           real_file_name, pos.line_number, (size_t)(-1), false,
+                           _("warning: syntax error"));
+      for (; c != UEOF && !is_quotable (c); c = phase4_getc (catr))
         {
           if (buflen >= bufmax)
             {
@@ -826,13 +827,14 @@ read_string (lex_pos_ty *pos)
 /* Read a .strings file from a stream, and dispatch to the various
    abstract_catalog_reader_class_ty methods.  */
 static void
-stringtable_parse (abstract_catalog_reader_ty *pop, FILE *file,
-                   const char *real_filename, const char *logical_filename)
+stringtable_parse (abstract_catalog_reader_ty *catr, FILE *file,
+                   const char *real_filename, const char *logical_filename,
+                   bool is_pot_role)
 {
   fp = file;
   real_file_name = real_filename;
-  gram_pos.file_name = xstrdup (real_file_name);
-  gram_pos.line_number = 1;
+  pos.file_name = xstrdup (real_file_name);
+  pos.line_number = 1;
   encoding = enc_undetermined;
   expect_fuzzy_msgstr_as_c_comment = false;
   expect_fuzzy_msgstr_as_cxx_comment = false;
@@ -852,23 +854,24 @@ stringtable_parse (abstract_catalog_reader_ty *pop, FILE *file,
       fuzzy_msgstr = NULL;
 
       /* Read the key and all the comments preceding it.  */
-      msgid = read_string (&msgid_pos);
+      msgid = read_string (catr, &msgid_pos);
       if (msgid == NULL)
         break;
 
-      special_comment_finish ();
+      special_comment_finish (catr);
 
       /* Skip whitespace.  */
       do
-        c = phase4_getc ();
+        c = phase4_getc (catr);
       while (is_whitespace (c));
 
       /* Expect a '=' or ';'.  */
       if (c == UEOF)
         {
-          po_xerror (PO_SEVERITY_ERROR, NULL,
-                     real_file_name, gram_pos.line_number, (size_t)(-1), false,
-                     _("warning: unterminated key/value pair"));
+          catr->xeh->xerror (CAT_SEVERITY_ERROR, NULL,
+                             real_file_name, pos.line_number, (size_t)(-1),
+                             false,
+                             _("warning: unterminated key/value pair"));
           break;
         }
       if (c == ';')
@@ -877,20 +880,22 @@ stringtable_parse (abstract_catalog_reader_ty *pop, FILE *file,
              necessarily designate an untranslated entry.  */
           msgstr = xstrdup ("");
           msgstr_pos = msgid_pos;
-          po_callback_message (NULL, msgid, &msgid_pos, NULL,
-                               msgstr, strlen (msgstr) + 1, &msgstr_pos,
-                               NULL, NULL, NULL,
-                               false, next_is_obsolete);
+          catalog_reader_seen_message (catr,
+                                       NULL, msgid, &msgid_pos, NULL,
+                                       msgstr, strlen (msgstr) + 1, &msgstr_pos,
+                                       NULL, NULL, NULL,
+                                       false, next_is_obsolete);
         }
       else if (c == '=')
         {
           /* Read the value.  */
-          msgstr = read_string (&msgstr_pos);
+          msgstr = read_string (catr, &msgstr_pos);
           if (msgstr == NULL)
             {
-              po_xerror (PO_SEVERITY_ERROR, NULL,
-                         real_file_name, gram_pos.line_number, (size_t)(-1),
-                         false, _("warning: unterminated key/value pair"));
+              catr->xeh->xerror (CAT_SEVERITY_ERROR, NULL,
+                                 real_file_name, pos.line_number, (size_t)(-1),
+                                 false,
+                                 _("warning: unterminated key/value pair"));
               break;
             }
 
@@ -899,7 +904,7 @@ stringtable_parse (abstract_catalog_reader_ty *pop, FILE *file,
           expect_fuzzy_msgstr_as_c_comment = next_is_fuzzy;
           do
             {
-              c = phase4_getc ();
+              c = phase4_getc (catr);
               if (fuzzy_msgstr != NULL)
                 expect_fuzzy_msgstr_as_c_comment = false;
             }
@@ -914,12 +919,12 @@ stringtable_parse (abstract_catalog_reader_ty *pop, FILE *file,
               if (fuzzy_msgstr == NULL && next_is_fuzzy)
                 {
                   do
-                    c = phase3_getc ();
+                    c = phase3_getc (catr);
                   while (c == ' ');
                   phase3_ungetc (c);
 
                   expect_fuzzy_msgstr_as_cxx_comment = true;
-                  c = phase4_getc ();
+                  c = phase4_getc (catr);
                   phase4_ungetc (c);
                   expect_fuzzy_msgstr_as_cxx_comment = false;
                 }
@@ -927,32 +932,34 @@ stringtable_parse (abstract_catalog_reader_ty *pop, FILE *file,
                 msgstr = fuzzy_msgstr;
 
               /* A key/value pair.  */
-              po_callback_message (NULL, msgid, &msgid_pos, NULL,
-                                   msgstr, strlen (msgstr) + 1, &msgstr_pos,
-                                   NULL, NULL, NULL,
-                                   false, next_is_obsolete);
+              catalog_reader_seen_message (catr,
+                                           NULL, msgid, &msgid_pos, NULL,
+                                           msgstr, strlen (msgstr) + 1, &msgstr_pos,
+                                           NULL, NULL, NULL,
+                                           false, next_is_obsolete);
             }
           else
             {
-              po_xerror (PO_SEVERITY_ERROR, NULL,
-                         real_file_name, gram_pos.line_number, (size_t)(-1),
-                         false,
-                         _("warning: syntax error, expected ';' after string"));
+              catr->xeh->xerror (CAT_SEVERITY_ERROR, NULL,
+                                 real_file_name, pos.line_number, (size_t)(-1),
+                                 false,
+                                 _("warning: syntax error, expected ';' after string"));
               break;
             }
         }
       else
         {
-          po_xerror (PO_SEVERITY_ERROR, NULL,
-                     real_file_name, gram_pos.line_number, (size_t)(-1), false,
-                     _("warning: syntax error, expected '=' or ';' after string"));
+          catr->xeh->xerror (CAT_SEVERITY_ERROR, NULL,
+                             real_file_name, pos.line_number, (size_t)(-1),
+                             false,
+                             _("warning: syntax error, expected '=' or ';' after string"));
           break;
         }
     }
 
   fp = NULL;
   real_file_name = NULL;
-  gram_pos.line_number = 0;
+  pos.line_number = 0;
 }
 
 const struct catalog_input_format input_format_stringtable =
